import {file_url, is_mac, system_file, zip_file, nodes} from "../nodes.js";

import {NaiveUiResolver} from "unplugin-vue-components/dist/resolvers.js";

import route_vue_set from "alone-frame/route/setVue.js";

import viteCompression from "vite-plugin-compression";

import Components from 'unplugin-vue-components/vite';

import AutoImport from "unplugin-auto-import/vite";

import Markdown from 'unplugin-vue-markdown/vite';

import {createHtmlPlugin} from "vite-plugin-html";

import {viteMockServe} from 'vite-plugin-mock';

import svgLoader from 'vite-svg-loader';

import vue from "@vitejs/plugin-vue";

import alone from "../alone.js";

import path from "path";

export const vueConfig = async function (pathname, option = {}) {
    const self = {
        //配置
        option: Object.assign({
            //模式
            mode: "development",
            //打包还是服务
            command: "build",
            //临时信息
            cache: "file/cache.json",
            //配置文件名
            conf: "config/conf.js",
            //js文件名
            env: "config/env.js",
            //env.js开发key
            dev: "_dev_",
            //App.vue文件名
            app: "App.vue",
            //main.js文件名
            main: "main.js",
            //当前目录
            dirname: file_url(import.meta.url, 2),
            //复制的目录
            copy_dir: alone.rtrim(file_url(import.meta.url).replace(/\\/g, '/'), '/') + "/src",
            //根目录
            pathname: process.cwd(),
            //开发目录
            path: (pathname || "src"),
        }, option || {}),
        //判断是否/开头
        is_run(path) {
            return path.substring(0, 1) == '/';
        },
        //文件格式
        format: {
            img: ['jpeg', 'jpg', 'png', 'gif', 'svg', 'svgz', 'bmp', 'tiff', 'webp', 'heif', 'heic', 'ico'],
            font: ['ttf', 'otf', 'woff', 'woff2', 'eot'],
            audio: ['mp3', 'wav', 'ogg', 'aac', 'flac'],
            video: ['mp4', 'webm', 'ogv', 'avi', 'mov']
        },
        //打包配置
        vendor_config: {
            //格式 包含有'svg', 'svgz',生成vendor.svg, svg或者file体积太大时可以选择开启
            svg: ['svg', 'svgz'],
            //路径包含有/vue-,生成vendor.vue, chunk体积太大时可以选择开启
            vue: (file, format) => {
                if (file.includes('/vue-') && format.includes('js')) {
                    return true;
                }
            },
            //路径包含有/alone/,生成vendor.alone
            alone: (file, format) => {
                if (file.includes('/alone') && format.includes('js')) {
                    return true;
                }
            },
            //路径包含有/naive-ui,生成vendor.naive, 推荐开启
            naive: (file, format) => {
                if (file.includes('/naive-ui') && format.includes('js')) {
                    return true;
                }
            },
            //格式 包含图片,字体,音频,视频打包成 vendor.file, 推荐开启
            file: (file, format) => {
                if (self.format_arr.includes(format)) {
                    return true;
                }
            },
            //路径包含有node_modules helper 打包成vendor.chunk, 推荐开启
            chunk: (file, format) => {
                const chunk = ['node_modules', 'helpers', '-helper'];
                if (chunk.some(val => file.includes(val))) {
                    return true;
                }
            }
        },
        //默认env
        env: {},
        //加载文件
        async import_file(file) {
            return await import(system_file(file));
        },
        //jsconfig.json配置
        js_config(object, path, pub, view, paths = {}) {
            paths["~/*"] = ["./*"];
            paths["&/*"] = ["./" + alone.trim(pub, '/') + "/*", "./" + path + "/" + alone.trim(pub, '/') + "/*"];
            paths["%/*"] = ["./" + path + "/" + view + "/*"];
            paths["@/*"] = ["./" + path + "/*"];
            object.compilerOptions = object.compilerOptions || {};
            object.compilerOptions.paths = paths;
            object.include = ["./" + path + "/**/*"];
            return object;
        },
        //启动
        async start() {
            this.update = false;
            //是否开发模式
            this.mode = this.option.mode == 'development';
            //当前目录node
            this.vite = nodes(this.option.dirname);
            //根目录node
            this.node = nodes(this.option.pathname);
            //复制目录node
            this.copy = nodes(this.option.copy_dir);
            //开发目录
            this.app_path = alone.trim(this.option.path.replace(/\\/g, '/'), '/');
            //开发目录node
            this.app = nodes(this.node.get_path(this.app_path));
            //复制env
            this.option.env = alone.trim(this.option.env.replace(/\\/g, '/'), '/');
            if (!this.app.is_exist(this.option.env)) {
                this.app.cp_file(this.copy.get_path("config/env.js"), this.option.env, false);
            }
            //env内容
            this.conf_env = (await this.import_file(this.app.get_path(this.option.env))).default || {};
            //处理env
            Object.keys(this.conf_env).forEach((k) => {
                if (this.mode) {
                    const key = this.option.dev + k;
                    if (this.conf_env.hasOwnProperty(key)) {
                        this.env[k] = this.conf_env[key];
                    } else if (k.substring(0, this.option.dev.length) == this.option.dev) {
                        this.env[k.substring(this.option.dev.length)] = this.conf_env[k];
                    } else if (!this.env.hasOwnProperty(k)) {
                        this.env[k] = this.conf_env[k];
                    }
                } else if (k.substring(0, this.option.dev.length) != this.option.dev) {
                    this.env[k] = this.conf_env[k];
                }
            });
            //视图目录
            this.env.view = alone.trim(this.env.view || 'view', '/');
            //静态目录
            this.env.public = alone.trim((this.env.public || 'public').replace(/\\/g, '/'), '/');
            //打包配置
            this.env.build = this.env.build || {};
            //临时数据
            this.cache = this.node.get_json(this.option.cache);
            if (!alone.isset(this.cache) || this.cache.path != this.app_path || this.cache.view != this.env.view) {
                this.update = true;
                this.vite.save_json(this.option.cache, {path: this.app_path, view: this.env.view});
            }
            //jsconfig.json
            this.is_jsconfig = this.node.is_exist('jsconfig.json');
            if (this.update || !this.is_jsconfig) {
                this.jsconfig_object = !this.is_jsconfig ? this.vite.get_json('jsconfig.json') : this.node.get_json('jsconfig.json');
                this.jsconfig_object = this.js_config(this.jsconfig_object, this.app_path, this.env.public, this.env.view);
                this.node.save_json('jsconfig.json', this.jsconfig_object);
            }
            //tsconfig.json
            this.is_tsconfig = this.node.is_exist('tsconfig.json');
            if (!this.is_tsconfig) {
                this.node.save_json('tsconfig.json', this.vite.get_json('tsconfig.json'));
            }
            //App.vue不存在时
            if (!this.app.is_exist(this.option.app)) {
                this.update = true;
                this.app.cp_file(this.copy.get_path("App.vue"), this.option.app, false);
                this.app.cp_file(this.copy.get_path("main.js"), this.option.main, false);
                this.app.cp_file(this.copy.get_path("config/conf.js"), this.option.conf, false);
                this.app.cp_dir_file(this.copy.get_path("view"), this.env.view, [], false);
                this.app.cp_dir_file(this.copy.get_path("mock"), 'mock', [], false);
                this.app.cp_dir_file(this.copy.get_path("style"), 'style', [], false);
            }
            //复制 layui
            if (!this.node.is_exist(this.env.public + "/layui/layui.js")) {
                this.node.cp_dir_file(this.vite.get_path("style/layui"), this.env.public + "/layui", [], false);
                this.node.cp_file(this.vite.get_path("style/lay.js"), this.env.public + "/layui/lay.js", false);
            }
            //项目public复制到public
            if (this.app.is_exist(this.env.public)) {
                this.node.cp_dir_file(this.app.get_path(this.env.public), this.env.public, false);
            }
            //mock目录
            this.env.mock = (this.env.mock === true && this.mode) ? this.app.get_path('mock') : "";
            //打包目录
            if (alone.isset(this.env.dist)) {
                this.out_dir = this.is_run(this.env.dist) ? this.env.dist : alone.trim(alone.trim(this.env.dist, '.'), '/');
            } else {
                this.dist = "dist/" + (this.app_path.split('/').length > 1 ? (this.app_path.split('/').slice(1).join('/')) : this.app_path);
            }
            if (!this.out_dir) {
                this.out_dir = this.node.get_path(this.dist);
            }
            //favicon.ico
            if (this.env.favicon && !this.is_run(this.env.favicon)) {
                this.env.favicon = alone.trim(alone.trim(this.env.favicon, '.'), '/');
                this.env.favicon = this.app.is_exist(this.env.favicon) ? (this.app_path + "/" + this.env.favicon) : "";
            }
            //conf内容
            this.config_conf = (this.app.is_exist(this.option.conf) ? ((await this.import_file(this.app.get_path(this.option.conf))).default) : {}) || {};
            //处理env
            this.del_env = ['host', 'port', 'build', 'zip', 'public', 'mock', 'favicon', 'view', 'dist', 'layui'];
            this.config_env = {...this.env};
            this.del_env.forEach((k) => ((this.config_env.hasOwnProperty(k)) && delete this.config_env[k]));
            //文件格式 array
            this.format_arr = [].concat(...Object.values(this.format));
            //打包配置
            this.chunk_config = {svg: false, vue: false, alone: true, naive: true, file: true, chunk: true};
            Object.keys(this.vendor_config).forEach((k) => {
                if (this.env.build[k] === true) {
                    this.chunk_config[k] = this.vendor_config[k];
                } else if (alone.isset(this.env.build[k])) {
                    this.chunk_config[k] = this.env.build[k];
                } else if (!this.env.build.hasOwnProperty(k) && this.chunk_config[k] === true) {
                    this.chunk_config[k] = this.vendor_config[k];
                } else {
                    delete this.chunk_config[k];
                }
            });
            if (alone.isset(this.env.build.callback) && alone.is_object(this.env.build.callback)) {
                Object.keys(this.env.build.callback).forEach((key) => {
                    if (alone.isset(this.env.build.callback[key])) {
                        this.chunk_config[key] = this.env.build.callback[key];
                    }
                });
            }
            //生成路由
            route_vue_set(this.app.get_path(this.env.view), {run: true});
            this.html_tag();
            return this;
        },
        html_tag() {
            this.tag_arr = [
                {
                    tag: 'meta',
                    attrs: {charset: 'UTF-8'},
                    injectTo: 'head',
                },
                {
                    tag: 'meta',
                    attrs: {
                        name: 'viewport',
                        content: 'width=device-width, initial-scale=1.0',
                    },
                    injectTo: 'head',
                },
                {
                    tag: 'div',
                    attrs: {id: "app", class: "app_main"},
                    injectTo: 'body',
                    children: `<style>body{margin:0}.sk-flow{width:50px;height:50px;display:flex;justify-content:space-between}.sk-flow-dot{width:25%;height:25%;background-color:#79bbff;border-radius:50%;animation:sk-flow 1.4s cubic-bezier(.455,.03,.515,.955)0s infinite both}.sk-flow-dot:first-child{animation-delay:-.3s}.sk-flow-dot:nth-child(2){animation-delay:-.15s}.sk-main{width:100vw;height:100vh;display:flex;align-items:center;justify-content:center}@keyframes sk-flow{0%,100%,80%{transform:scale(.3)}40%{transform:scale(1)}}</style><div class="sk-main"><div class="sk-flow"><div class="sk-flow-dot"></div><div class="sk-flow-dot"></div><div class="sk-flow-dot"></div></div></div>`,
                }
            ];
            this.lay_def = {status: true, link: "/layui/css/layui.css", script: "/layui/layui.js"};
            this.env.layui = this.env.layui || this.lay_def;
            if (this.env.layui.status !== false) {
                this.tag_arr = this.tag_arr.concat([
                    {
                        tag: 'link',
                        attrs: {rel: 'stylesheet', href: this.env.layui.link || this.lay_def.link},
                        injectTo: 'head',
                    },
                    {
                        tag: 'script',
                        attrs: {type: 'module', src: this.env.layui.script || this.lay_def.script},
                        injectTo: 'head-prepend',
                    },
                    {
                        tag: 'script',
                        attrs: {type: 'module'},
                        injectTo: 'head',
                        children: `window.$ = layui.$;`,
                    },
                ]);
            }
            if (this.env.favicon) {
                this.tag_arr = this.tag_arr.concat([{
                    tag: 'link',
                    attrs: {icon: 'favicon', rel: "icon", href: this.env.favicon},
                    injectTo: 'head',
                }]);
            }
        },
        publicDir() {
            return this.node.get_path(this.env.public);
        },
        define() {
            return {'process.env': this.config_env, 'process.config': this.config_conf};
        },
        resolve() {
            return {
                alias: {
                    '~': this.node.get_path('.'),
                    '&': this.node.get_path(this.env.public),
                    '@': this.app.get_path('.'),
                    '%': this.app.get_path(this.env.view)
                }
            };
        },
        preview() {
            return {host: this.env.host || '127.0.0.1', port: this.env.port || '8088'};
        },
        server() {
            return Object.assign({
                proxy: {
                    '/api': {
                        target: this.env.api,
                        changeOrigin: true,
                        rewrite: (path) => path.replace(/^\/api/, '')
                    }
                }
            }, this.preview());
        },
        build() {
            return {
                outDir: this.out_dir,
                target: 'esnext',//'esnext', es2015 es2019
                minify: 'terser',//'terser',esbuild
                cssCodeSplit: true, // 如果设置为false，整个项目中的所有 CSS 将被提取到一个 CSS 文件中
                reportCompressedSize: false,//  关闭生成map文件 可以达到缩小打包体积
                chunkSizeWarningLimit: 1500,
                sourcemap: false, // 这个生产环境一定要关闭，不然打包的产物会很大
                terserOptions: {
                    format: {
                        beautify: false, // 不格式化输出
                        comments: false, // 不保留注释
                    },
                    mangle: {
                        toplevel: true, // 混淆顶级变量和函数
                    },
                    compress: {
                        properties: true,
                        dead_code: true,    // 删除未使用的代码
                        drop_console: true,//打包时删除console
                        drop_debugger: true,//打包时删除 debugger
                    },
                    sourceMap: true, // 生成源映射文件
                    ecma: 2015,      // 使用 ECMAScript 2015 的特性
                    enclose: false,  // 不包裹代码在闭包中
                    keep_classnames: false, // 不保留类名
                    keep_fnames: false, // 不保留函数名
                    ie8: false,       // 不兼容 IE8
                    module: false,    // 不支持 ES6 模块语法
                    nameCache: null,  // 不使用名称缓存
                    safari10: false,  // 不兼容 Safari 10
                    toplevel: false,  // 不混淆顶级作用域的变量
                },
                rollupOptions: {
                    treeshake: true, // 开启 Tree Shaking，消除未使用的代码，减小最终的包大小
                    output: {
                        hashCharacters: 'hex',
                        chunkFileNames: (obj) => {
                            const name = (obj.name).toString().toLowerCase()
                            return 'assets/js/' + name + '.[hash].js';
                        },
                        entryFileNames: (obj) => {
                            const name = (obj.name).toString().toLowerCase()
                            return 'assets/js/' + name + '.[hash].js';
                        },
                        assetFileNames: (obj) => {
                            const type = obj.type.toString().toLowerCase();
                            if (type === 'asset' && (obj.name || obj.names)) {
                                const file = (obj.name || (obj.names.slice(0, 1).join(''))).toString();
                                const files = file.replace(/\\/g, '/').toLowerCase();
                                const format = path.extname(files).slice(1);
                                const array = Object.keys(this.format);
                                for (const key of array) {
                                    const val = this.format[key];
                                    if (val && ((typeof val == 'function') ? val(files, format) : (Array.isArray(val) ? val.includes(format) : files.includes(val)))) {
                                        return 'assets/' + key + '/[name].[hash].[ext]';
                                    }
                                }
                            }
                            return 'assets/[ext]/[name].[hash].[ext]';
                        },
                        manualChunks: (file) => {
                            const files = file.replace(/\\/g, '/').toLowerCase();
                            const format = path.extname(files).slice(1);
                            const array = Object.keys(this.chunk_config);
                            for (const key of array) {
                                const val = this.chunk_config[key];
                                if (val && ((typeof val == 'function') ? val(files, format) : (Array.isArray(val) ? val.includes(format) : files.includes(val)))) {
                                    return "vendor." + key;
                                }
                            }
                        }
                    }
                }
            };
        }
    };
    return (await self.start());
}

export default async function (pathname, option = {}) {
    const self = await vueConfig(pathname, option);
    const plugins = {
        vue(option = {}) {
            return vue(Object.assign({include: [/\.vue$/, /\.md$/]}, option || {}));
        },
        svgLoader(option = {}) {
            return svgLoader(Object.assign({defaultImport: 'url'}, option || {}));
        },
        viteMockServe(option = {}) {
            return (alone.isset(self.env.mock) ? viteMockServe(Object.assign({
                watchFiles: true,//指示是否监视 Mock 文件的变化。如果设置为 true，当 Mock 文件发生变化时，Mock 服务器会自动更新。
                enable: true,//指示是否启用 Mock 功能。设置为 false 可以临时禁用 Mock。
                logger: false,//指示是否启用插件的日志记录。设置为 true 会在终端输出 Mock 请求的日志信息。
                cors: true //指示是否启用 CORS（跨域资源共享）。设置为 true 会在响应中添加 CORS 头部，以允许跨域请求。
            }, option || {}, {
                mockPath: self.env.mock,//存放 Mock 文件的路径。默认情况下，插件会在此路径下查找 Mock 文件
            })) : {});
        },
        Markdown(option = {}) {
            return Markdown(Object.assign({
                headEnabled: true,
                include: [`${self.node.get_path('.')}/*.md`, `${self.node.get_path('.')}/**/*.md`],
                markdownItOptions: {html: true, linkify: true, typographer: true}
            }, option || {}));
        },
        customBuild() {
            return {
                name: 'custom-build',
                buildStart: () => {
                    console.log('path', `${is_mac() ? "file://" : ""}${self.app.get_path('.')}`);
                },
                closeBundle: () => {
                    console.log('complete', `${is_mac() ? "file://" : ""}` + self.out_dir);
                    if (self.option.command == 'build') {
                        if (self.env.zip === true) {
                            zip_file(self.out_dir, alone.rtrim(self.out_dir, '/') + ".zip", function (size, unit, file) {
                                console.log('zip file', file);
                                console.log('zip size', `${size}${unit}`);
                            });
                        }
                    }
                }
            }
        },
        createHtmlPlugin(option = {}) {
            option = Object.assign(option || {}, {
                minify: true,
                entry: self.app_path + "/" + self.option.main,
                inject: {tags: []}
            });
            option.inject.tags = option.inject.tags.concat(self.tag_arr);
            return createHtmlPlugin(option);
        },
        viteCompression(option = {}) {
            return viteCompression(Object.assign({
                verbose: true,    // 默认即可
                disable: true,  //开启压缩(不禁用)，默认即可
                deleteOriginFile: false, //删除源文件
                threshold: 10240, //压缩前最小文件大小
                algorithm: 'gzip',  //压缩算法
                ext: '.gz', //文件类型
            }, option || {}));
        },
        AutoImport(option = {}) {
            return AutoImport(Object.assign({
                imports: ['vue', {
                    'naive-ui': [
                        'useDialog',
                        'useMessage',
                        'useNotification',
                        'useLoadingBar'
                    ]
                }]
            }, option || {}));
        },
        Components(option = {}) {
            return Components(Object.assign({
                extensions: ['vue', 'md'],
                include: [/\.vue$/, /\.vue\?vue/, /\.md$/],
                resolvers: [NaiveUiResolver()],
            }, option || {}))
        }
    };
    return {
        publicDir: self.publicDir(),
        define: self.define(),
        server: self.server(),
        preview: self.preview(),
        resolve: self.resolve(),
        build: self.build(),
        plugin: plugins,
        plugins: [
            plugins.vue(),
            plugins.svgLoader(),
            plugins.viteMockServe(),
            plugins.Markdown(),
            plugins.customBuild(),
            plugins.createHtmlPlugin(),
            plugins.viteCompression(),
            plugins.AutoImport(),
            plugins.Components()
        ]
    };
}